@@ -12,6 +12,7 @@ from keras_preprocessing.image import load_img
 from pointgrid import align_points_to_grid
 from distutils.dir_util import copy_tree
 from sklearn.decomposition import PCA
+from scipy.spatial import ConvexHull
 from iiif_downloader import Manifest
 from collections import defaultdict
 from rasterfairy import coonswarp
@@ -836,30 +837,53 @@ def get_centroids(**kwargs):
   '''Return the stable clusters from the condensed tree of connected components from the density graph'''
   print(' * HDBSCAN clustering data with ' + str(multiprocessing.cpu_count()) + ' cores...')
   config = {
+    'core_dist_n_jobs': multiprocessing.cpu_count(),
     'min_cluster_size': kwargs['min_cluster_size'],
     'cluster_selection_epsilon': 0.01,
     'min_samples': 1,
-    'core_dist_n_jobs': multiprocessing.cpu_count(),
   }
-  z = HDBSCAN(**config).fit(kwargs['vecs'])
-  # find the centroids for each cluster
+  v = kwargs['vecs']
+  z = HDBSCAN(**config).fit(v)
+  # find the points in each cluster
   d = defaultdict(list)
   for idx, i in enumerate(z.labels_):
-    d[i].append(kwargs['vecs'][idx])
+    d[i].append(v[idx])
+  # find the convex hull for each cluster's points
+  convex_hulls = []
+  for i in d:
+    hull = ConvexHull(d[i])
+    points = [hull.points[j] for j in hull.vertices]
+    # the last convex hull simplex needs to connect back to the first point
+    convex_hulls.append(np.vstack([points, points[0]]))
+  # find the centroids for each cluster
   centroids = []
   for i in d:
     x, y = np.array(d[i]).T
-    centroids.append(np.array([np.sum(x)/len(x), np.sum(y)/len(y)]))
-  closest, _ = pairwise_distances_argmin_min(centroids, kwargs['vecs'])
-  closest = set(closest)
-  print(' * found', len(closest), 'clusters')
+    centroids.append(np.array([np.mean(x), np.mean(y)]))
+  # identify the number of points in each cluster
+  lens = [len(d[i]) for i in d]
+  # combine data into cluster objects
+  closest, _ = pairwise_distances_argmin_min(centroids, v)
   paths = [kwargs['image_paths'][i] for i in closest]
-  data = [{
+  clusters = [{
     'img': clean_filename(paths[idx]),
-    'label': 'Cluster {}'.format(idx+1),
+    'convex_hull': convex_hulls[idx].tolist(),
+    'n_images': lens[idx],
   } for idx, i in enumerate(closest)]
+  # remove massive clusters
+  retained = []
+  for idx, i in enumerate(clusters):
+    x, y = np.array(i['convex_hull']).T
+    area = 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))
+    if area < 0.2:
+      retained.append(i)
+  # sort the clusers by size
+  clusters = sorted(retained, key=lambda i: i['n_images'], reverse=True)
+  for idx, i in enumerate(clusters):
+    i['label'] = 'Cluster {}'.format(idx+1)
   # save the centroids to disk and return the path to the saved json
-  return write_json(get_path('centroids', 'centroid', **kwargs), data, **kwargs)
+  print(' * found', len(clusters), 'clusters')
+  return write_json(get_path('centroids', 'centroid', **kwargs), clusters, **kwargs)
 
 
 def get_heightmap(path, label, **kwargs):
